﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Net;
using System.Text;
using System.Text.RegularExpressions;
using System.Xml;

class jQueryExternGenerator
{
    static void Main(string[] args)
    {
        string MaxApiVersion = "9.9.9.9";
        if (args.Length >= 1)
            MaxApiVersion = args[0];
        new jQueryExternGenerator(MaxApiVersion);
    }

    //Map types referenced in the jQuery API to Closure recognized types
    private Dictionary<string, string> TypesMap = new Dictionary<string, string>(StringComparer.CurrentCultureIgnoreCase);

    //Map jQuery properties and method return types that use "Object" to stricter types
    private Dictionary<string, string> ObjectTypes = new Dictionary<string, string>(StringComparer.CurrentCultureIgnoreCase);

    //Map jQuery properties and method return types that use "Array" to stricter types
    private Dictionary<string, string> ArrayTypes = new Dictionary<string, string>(StringComparer.CurrentCultureIgnoreCase);

    //Map of all jQuery methods and properties
    private Dictionary<string, string> Entries = new Dictionary<string, string>();

    //Methods which belong to the following categories will be given the @nosideeffect annotation
    private List<string> NoSideEffectCategories = new List<string>();

    //Methods which begin with the following strings will be given the @nosideeffect annotation
    private List<string> NoSideEffectPrefixes = new List<string>();

    //The following methods will be given the @nosideeffect annotation
    private Dictionary<string, string> NoSideEffectMethods = new Dictionary<string, string>(StringComparer.CurrentCultureIgnoreCase);

    private jQueryExternGenerator(string MaxApiVersion)
    {
        InitializeTypesMap(MaxApiVersion);
        InitializeArrayTypes(MaxApiVersion);
        InitializeObjectTypes(MaxApiVersion);
        InitializeEntries(MaxApiVersion);
        InitializeNoSideEffectMaps(MaxApiVersion);

        //Map jQuery parameter names which are JavaScript reserved words to alternate names
        Dictionary<string, string> ReservedNames = new Dictionary<string, string>(StringComparer.CurrentCultureIgnoreCase);
        ReservedNames.Add("Object", "obj");
        ReservedNames.Add("Array", "arr");
        ReservedNames.Add("Element", "elem");
        ReservedNames.Add("function", "fnc");
        ReservedNames.Add("switch", "flag");
        ReservedNames.Add("arguments", "args");

        // Reformat single JSDoc annotations to occupy only one line and wrap at 80 characters
        Regex OneLineJSDocTagRegEx = new Regex(@"/\*\*\n \*\s*(.+)\n \*/", RegexOptions.IgnoreCase | RegexOptions.Compiled);
        Regex LongLineJSDocRegEx = new Regex(@"([^\n]{81,})(\n|$)", RegexOptions.Compiled | RegexOptions.Multiline);
        Regex jQueryDeclerationRegEx = new Regex(@"^jQuery\.", RegexOptions.Compiled | RegexOptions.Multiline);

        string ApiVersion = "0.0.0.0";

        //Load the jQuery API Xml
        XmlDocument jQueryXml = new XmlDocument();
        jQueryXml.Load("http://api.jquery.com/api/");

        //Select and loop through all of the method nodes
        XmlNodeList NodeList = jQueryXml.SelectNodes("/api/entries/entry[@type='method']");
        int PositionIndex = 0;
        foreach (XmlNode Method in NodeList)
        {
            string EntryName = Method.Attributes["name"].Value;
            if (Method.Attributes["name"].Value.Equals("jQuery"))
                EntryName = "jQuery";
            else if (Method.Attributes["name"].Value.StartsWith("jQuery."))
                EntryName = Method.Attributes["name"].Value;
            else if (Method.Attributes["name"].Value.Contains("."))
            {
                EntryName = Method.Attributes["name"].Value.Substring(0, Method.Attributes["name"].Value.IndexOf('.'));
                EntryName = "jQuery." + Method.Attributes["name"].Value.Replace(".", ".prototype.");
            }
            else
                EntryName = "jQuery.prototype." + EntryName;

            if (Entries.ContainsKey(EntryName))
                continue;

            StringBuilder OutputData = new StringBuilder();
            OutputData.Append("/**\n");

            //Since methods can have more than one entry (due to different return types), select all method nodes with the same name.
            XmlNodeList Methods = jQueryXml.SelectNodes("/api/entries/entry[@type='method' and @name='" + Method.Attributes["name"].Value + "']");

            //Keep track of all possible return types for this method.
            Dictionary<string, string> ReturnTypes = new Dictionary<string, string>();
            foreach (XmlNode Entry in Methods)
            {
                if (Entry.Attributes["return"] != null && !String.IsNullOrEmpty(Entry.Attributes["return"].Value))
                {
                    string EntryReturnType = LookupType(Entry.Attributes["return"].Value, EntryName);

                    if (!ReturnTypes.ContainsKey(EntryReturnType))
                        ReturnTypes.Add(EntryReturnType, EntryReturnType);
                }
            }

            //Select all of the method signatures
            XmlNodeList Signatures = jQueryXml.SelectNodes("/api/entries/entry[@type='method' and @name='" + Method.Attributes["name"].Value + "']/signature");

            //List of arguments for each signature
            List<List<jQueryEntryArgument>> SignatureArgs = new List<List<jQueryEntryArgument>>();

            //Note whether a method has a function type parameter. Used to netgate @nosideeffect annotations.
            bool hasFunctionParameter = false;

            //Process each signature
            foreach (XmlNode Signature in Signatures)
            {
                //Keep track of the newest version number seen
                XmlNode VersionAdded = Signature.SelectSingleNode("added");
                if (VersionAdded != null)
                {
                    if (VersionAdded.InnerText.CompareTo(MaxApiVersion) > 0)
                        continue;

                    if (VersionAdded.InnerText.CompareTo(ApiVersion) > 0)
                        ApiVersion = VersionAdded.InnerText;
                }

                //List of arguments for this signature
                List<jQueryEntryArgument> Args = new List<jQueryEntryArgument>();
                XmlNodeList SignatureArgumentNodes = Signature.SelectNodes("argument");
                PositionIndex = 1;
                foreach (XmlNode SignatureArgumentNode in SignatureArgumentNodes)
                {
                    jQueryEntryArgument Arg = new jQueryEntryArgument(PositionIndex);
                    if (SignatureArgumentNode.Attributes["optional"] != null && SignatureArgumentNode.Attributes["optional"].Value == "true")
                        Arg.Optional = true;

                    Arg.Name = SignatureArgumentNode.Attributes["name"].Value;

                    string TempArgType = LookupType(SignatureArgumentNode.Attributes["type"].Value, EntryName + "(" + Arg.Name + ")");
                    string[] ArgTypes = TempArgType.Split('|');
                    foreach (string ArgType in ArgTypes)
                    {
                        TempArgType = ArgType;

                        //If it is a function type, lookup the parameter types and trim the name
                        if (TempArgType.StartsWith("function") && Arg.Name.Contains("("))
                        {
                            TempArgType = LookupFunctionTypes(Arg.Name);
                            hasFunctionParameter = true;
                            Arg.Name = Arg.Name.Substring(0, Arg.Name.IndexOf("(")).Trim();
                        }
                        //Handle any type or variable number of argument cases
                        else if ((TempArgType == "*" || TempArgType == "Object") && Arg.Name == "args")
                            TempArgType = "...*";

                        Arg.AddType(TempArgType);
                    }
                    Args.Add(Arg);

                    PositionIndex++;
                }

                SignatureArgs.Add(Args);
            }

            //Keep track of the number of required arguments for ALL signatures
            int NumRequiredArgs = 999;
            foreach (List<jQueryEntryArgument> ArgList in SignatureArgs)
            {
                //Number of required arguments for this signature.
                int NumSigRequiredArgs = 0;
                for (int i = 0; i < ArgList.Count; i++)
                {
                    if (ArgList[i].Optional)
                    {
                        //If an argument is optional it can be omitted. The argument at this posision
                        //must support the types of all subsequent optional paramters as well as the first
                        //required parameter.
                        for (int j = i + 1; j < ArgList.Count; j++)
                        {
                            ArgList[i].UnionTypes(ArgList[j].Types);

                            if (!ArgList[j].Optional)
                                break;
                        }
                    }
                    else
                        NumSigRequiredArgs++;
                }
                NumRequiredArgs = Math.Min(NumRequiredArgs, NumSigRequiredArgs);

                for (int i = ArgList.Count - 1; i >= 0; i--)
                {
                    //Remark arguments as optional based so that required arguments are listed first in the parameter list.
                    ArgList[i].Optional = i >= NumRequiredArgs;
                }
            }

            int UniqueArgNameSuffix = 2;
            bool PositionHasArguments = true, HasVarArgs = false;

            List<jQueryEntryArgument> OutputArgs = new List<jQueryEntryArgument>();
            Dictionary<string, string> UniqueArgNames = new Dictionary<string, string>();
            PositionIndex = 0;

            //loop through each signature and combine all the arguments at each position
            while (PositionHasArguments && !HasVarArgs)
            {
                PositionHasArguments = false;
                jQueryEntryArgument OutputArg = new jQueryEntryArgument(PositionIndex);
                OutputArg.Name = null;
                OutputArg.Optional = false;
                bool GenericName = false;

                for (int i = 0; i < SignatureArgs.Count; i++)
                {
                    if (SignatureArgs[i].Count > PositionIndex)
                    {
                        PositionHasArguments = true;
                        if (OutputArg.Name == null)
                            OutputArg.Name = SignatureArgs[i][PositionIndex].Name;
                        else if (!GenericName && !OutputArg.Name.Equals(SignatureArgs[i][PositionIndex].Name, StringComparison.CurrentCultureIgnoreCase))
                        {
                            //Signatures have different name for the argument at the current posistion. Use a generic name.
                            GenericName = true;
                            OutputArg.Name = "arg" + (PositionIndex + 1).ToString();
                        }

                        if (PositionIndex >= NumRequiredArgs)
                            OutputArg.Optional = true;

                        OutputArg.UnionTypes(SignatureArgs[i][PositionIndex].Types);

                        //Handle variable arguments
                        if (OutputArg.Types.Count == 1 && OutputArg.Types[0] == "...*")
                        {
                            OutputArg.Optional = false;
                            HasVarArgs = true;
                            OutputArg.Name = "var_args";
                            GenericName = true;
                            break;
                        }
                    }
                }


                if (PositionHasArguments)
                {
                    if (!GenericName)
                    {
                        //Some jquery argument names use javascript reserved keywords.
                        //Lookup a replacement name.
                        if (ReservedNames.ContainsKey(OutputArg.Name))
                            OutputArg.Name = ReservedNames[OutputArg.Name];

                        //If this argument name has already been used for this function (yes it happens),
                        //add a suffix value to keep it unique
                        if (UniqueArgNames.ContainsKey(OutputArg.Name))
                        {
                            OutputArg.Name += UniqueArgNameSuffix.ToString();
                            UniqueArgNameSuffix++;
                        }
                    }

                    UniqueArgNames.Add(OutputArg.Name, OutputArg.Name);

                    //Add this argument to the final list
                    OutputArgs.Add(OutputArg);

                    OutputData.Append(" * @param {");
                    if (OutputArg.Types.Count > 1)
                        OutputData.Append("(");

                    for (int i = 0; i < OutputArg.Types.Count; i++)
                    {
                        if (i > 0)
                            OutputData.Append("|");

                        OutputData.Append(OutputArg.Types[i]);
                    }

                    if (OutputArg.Types.Count > 1)
                        OutputData.Append(")");

                    if (OutputArg.Optional)
                        OutputData.Append("=");

                    OutputData.Append("} " + OutputArg.Name + "\n");
                }

                PositionIndex++;
            }

            //Output the return type annotation for this method
            if (ReturnTypes.Count > 0)
            {
                string ReturnType = "";
                if (ReturnTypes.ContainsValue("*"))
                    ReturnType = "*";
                else
                {
                    foreach (string EntryType in ReturnTypes.Values)
                    {
                        if (ReturnType.Length > 0)
                            ReturnType += "|";
                        ReturnType += EntryType;
                    }
                }

                OutputData.Append(" * @return {");
                if (ReturnType.Contains("|"))
                    OutputData.Append("(");

                OutputData.Append(ReturnType);

                if (ReturnType.Contains("|"))
                    OutputData.Append(")");

                OutputData.Append("}\n");
            }

            //Check for and add the @nosideeffect annotation when appropriate
            bool NoSideEffects = NoSideEffectMethods.ContainsKey(EntryName);
            if (!NoSideEffects)
            {
                foreach (string Category in NoSideEffectCategories)
                {
                    if (Method.SelectSingleNode("category[@name='" + Category + "']") != null)
                    {
                        NoSideEffects = true;
                        break;
                    }
                }
            }

            if (!NoSideEffects)
            {
                foreach (string NoSideEffectPrefix in NoSideEffectPrefixes)
                {
                    string MethodName = EntryName;
                    if (EntryName.IndexOf('.') >= 0)
                        MethodName = MethodName.Substring(EntryName.LastIndexOf('.') + 1);

                    if (MethodName.StartsWith(NoSideEffectPrefix, StringComparison.CurrentCultureIgnoreCase))
                    {
                        NoSideEffects = true;
                        break;
                    }
                }
            }

            //Methods which have function type arguments may not be marked as @nosideeffects
            //as the function could alter something.
            if (NoSideEffects && !hasFunctionParameter)
                OutputData.Append(" * @nosideeffects\n");

            //If there are no annotations for this method, add an @return {undefined} annotation
            //If the compiler sees a method without any annotations, it assumes it takes a variable number of arguments and can return any type.
            if (OutputData.ToString().Equals("/**\n"))
            {
                OutputData.Remove(0, OutputData.Length);
                OutputData.Append("/** @return {undefined}");
            }

            //Finished annotations
            OutputData.Append(" */\n");

            //Output the method signature
            if (EntryName.Equals("jQuery"))
                OutputData.Append("function jQuery(");
            else if (EntryName.StartsWith("jQuery.prototype."))
                OutputData.Append(EntryName + " = function(");
            else if (EntryName.StartsWith("jQuery."))
                OutputData.Append(EntryName + " = function(");
            else
                Console.Error.WriteLine("Unhandled name - " + EntryName);

            bool isFirstArg = true;
            foreach (jQueryEntryArgument Arg in OutputArgs)
            {
                if (isFirstArg)
                    isFirstArg = false;
                else
                    OutputData.Append(", ");

                OutputData.Append(Arg.Name);
            }
            OutputData.Append(") {};\n");

            if (EntryName.Equals("jQuery"))
                Entries.Add("jQuery", OutputData.ToString().Replace("/**\n", "/**\n * @constructor\n"));
            else
            {
                Entries.Add(EntryName, OutputData.ToString());
                if (EntryName.StartsWith("jQuery.") && EntryName.IndexOf(".prototype.") < 0)
                    Entries.Add(EntryName + "$", jQueryDeclerationRegEx.Replace(OutputData.ToString(), @"$."));
            }
        }

        //Select and loop through all of the property nodes
        NodeList = jQueryXml.SelectNodes("/api/entries/entry[@type='property']");
        foreach (XmlNode Property in NodeList)
        {
            string EntryName = Property.Attributes["name"].Value;
            //Add the property definition
            if (Property.Attributes["name"].Value.StartsWith("jQuery."))
            {
            }
            else if (Property.Attributes["name"].Value.Contains("."))
                EntryName = "jQuery." + Property.Attributes["name"].Value.Replace(".", ".prototype.");
            else
                EntryName = "jQuery.prototype." + EntryName;

            XmlNodeList PropertySignatures = Property.SelectNodes("signature");
            if (PropertySignatures.Count > 1)
                Console.Error.WriteLine(Property.Attributes["name"].Value + " - property has more than one signature");

            XmlNode VersionAdded = Property.SelectSingleNode("signature/added");
            if (VersionAdded != null)
            {
                if (VersionAdded.InnerText.CompareTo(MaxApiVersion) > 0)
                    continue;

                if (VersionAdded.InnerText.CompareTo(ApiVersion) > 0)
                    ApiVersion = VersionAdded.InnerText;
            }

            if (Entries.ContainsKey(EntryName))
                continue;

            string EntryReturnType;
            if (Property.Attributes["return"] == null)
            {
                Console.WriteLine("Property " + EntryName + " does not specify a return type. Using Object.<string, *>");
                EntryReturnType = "Object.<string, *>";
            }
            else
                EntryReturnType = LookupType(Property.Attributes["return"].Value, EntryName);

            if (Property.Attributes["return"] == null)
            {
                Console.WriteLine("Property " + EntryName + " does not specify a return type. Using Object.<string, *>");
            }

            //Add the type annotation for the property
            string jsDoc = "/** @type {" + EntryReturnType + "} */\n";
            jsDoc += EntryName + ";\n";

            //Add the property to the list of entries.
            Entries.Add(EntryName, jsDoc);
            if (EntryName.StartsWith("jQuery.") && EntryName.IndexOf(".prototype.") < 0)
                Entries.Add(EntryName + "$", jQueryDeclerationRegEx.Replace(jsDoc, @"$."));
        }

        if (ApiVersion.CompareTo("1.6") < 0)
            Entries.Add("jQueryObject", "/** @typedef {jQuery} */\nvar jQueryObject;\n");

        string MajorApiVersion = ApiVersion;
        int SubversionIndex = ApiVersion.IndexOf('.');
        if (SubversionIndex > 0)
        {
            MajorApiVersion = ApiVersion.Substring(0, SubversionIndex);
            if (ApiVersion.Length > SubversionIndex + 1)
            {
                SubversionIndex = ApiVersion.IndexOf('.', SubversionIndex + 1);
                if (SubversionIndex > 0)
                    MajorApiVersion = ApiVersion.Substring(0, SubversionIndex);
                else
                    MajorApiVersion = ApiVersion;
            }
            else
                MajorApiVersion = ApiVersion;
        }

        string Filename = "jquery-" + MajorApiVersion + ".js";
        if (File.Exists(Filename))
            File.Delete(Filename);

        using (FileStream OutputFileStream = File.Create(Filename))
        using (StreamWriter OutputStream = new StreamWriter(OutputFileStream))
        {
            OutputStream.Write((@"/*
 * Copyright 2011 The Closure Compiler Authors.
 *
 * Licensed under the Apache License, Version 2.0 (the ""License"");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an ""AS IS"" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @fileoverview Externs for jQuery " + ApiVersion + @"
 *
 * Note that some functions use different return types depending on the number
 * of parameters passed in. In these cases, you may need to annotate the type
 * of the result in your code, so the JSCompiler understands which type you're
 * expecting. For example:
 *    <code>var elt = /** @type {Element} * / (foo.get(0));</code>
 *
 * @see http://api.jquery.com/
 * @externs
 */

").Replace("\r", string.Empty));
            List<jQueryEntry> SortedEntries = new List<jQueryEntry>();
            foreach (KeyValuePair<string, string> Entry in Entries)
            {
                if(Entry.Value != null)
                    SortedEntries.Add(new jQueryEntry(Entry.Key, Entry.Value));
            }
            SortedEntries.Sort();

            MatchEvaluator WrapLineEvaluator = new MatchEvaluator(WrapJSDocLine);
            for (int i = 0; i < SortedEntries.Count; i++)
            {
                string OutputEntry = SortedEntries[i].Value.Substring(0, SortedEntries[i].Value.Length - 1).Replace("\r", string.Empty);
                OutputEntry = LongLineJSDocRegEx.Replace(OutputEntry, WrapLineEvaluator);
                OutputEntry = OneLineJSDocTagRegEx.Replace(OutputEntry, "/** $1 */");
                OutputStream.Write(OutputEntry);
                if (i + 1 < SortedEntries.Count)
                    OutputStream.Write("\n\n");
            }
        }

        Console.WriteLine("done");
        Console.Read();
    }

    //Function takes a jQuery type and returns a Closure compatible type
    private string LookupType(string Input, string EntryName)
    {
        string ReturnType = string.Empty;

        if (Input.Length == 0)
            return "undefined";

        string[] InputTypes = Input.Split(',');
        if (TypesMap.ContainsKey(Input))
            InputTypes = new string[] { Input };

        for (int i = 0; i < InputTypes.Length; i++)
        {
            if (i > 0)
                ReturnType += "|";

            if (!TypesMap.ContainsKey(InputTypes[i].Trim()))
            {
                //jQuery documentation references a type we haven't explicitly accounted for
                //Warn the user then continue
                Console.Error.WriteLine("Unknown type: " + InputTypes[i]);
            }
            else
            {
                string tmpReturnType = TypesMap[InputTypes[i].Trim()];
                if (EntryName.Length > 0 && (tmpReturnType == "Object" || tmpReturnType == "Array"))
                {
                    if (tmpReturnType == "Object" && ObjectTypes.ContainsKey(EntryName))
                        tmpReturnType = ObjectTypes[EntryName];
                    else if (tmpReturnType == "Array" && ArrayTypes.ContainsKey(EntryName))
                        tmpReturnType = ArrayTypes[EntryName];
                    else
                    {
                        string ParamName = EntryName;
                        if (ParamName.IndexOf('(') >= 0)
                        {
                            ParamName = ParamName.Substring(EntryName.IndexOf('('));
                            if (tmpReturnType == "Object" && ObjectTypes.ContainsKey(ParamName))
                                tmpReturnType = ObjectTypes[ParamName];
                            else if (tmpReturnType == "Array" && ArrayTypes.ContainsKey(ParamName))
                                tmpReturnType = ArrayTypes[ParamName];
                        }


                        if (tmpReturnType == "Object" || tmpReturnType == "Array")
                            Console.WriteLine(EntryName + " " + Input + " uses Object type");
                    }
                }

                ReturnType += tmpReturnType;
            }
        }

        return ReturnType;
    }

    //Function takes a jQuery callback function description and returns a function definition with Closure compatible types
    private string LookupFunctionTypes(string FunctionName)
    {
        //Get the list of arguments
        string Types = FunctionName.Substring(FunctionName.IndexOf('(') + 1).Trim();
        Types = Types.Substring(0, Types.Length - 1);
        if (Types.Trim().Length == 0)
            return "function()";

        //Lookup the type of each function argument
        string[] FuncArgTypes = Types.Split(',');
        string ReturnVal = "function(";
        for (int i = 0; i < FuncArgTypes.Length; i++)
        {
            if (i > 0)
                ReturnVal += ",";

            string FuncArgType = LookupType(FuncArgTypes[i].Trim(), string.Empty);

            if (FuncArgType.Contains("|"))
                ReturnVal += "(" + FuncArgType + ")";
            else
                ReturnVal += FuncArgType;
        }

        ReturnVal += ")";

        return ReturnVal;
    }

    private string WrapJSDocLine(Match m)
    {
        string x = m.Groups[1].Value;
        if (x.StartsWith("/** ") || x.StartsWith(" * "))
        {
            string y = x.Substring(0, 80);
            int LineBreakPoint = y.LastIndexOf('|');
            if (LineBreakPoint >= 0)
            {
                if (x.IndexOf('>', LineBreakPoint) >= 0 && x.IndexOf('<', LineBreakPoint) < 0)
                    LineBreakPoint = y.Substring(0, LineBreakPoint - 1).LastIndexOf('|');

                x = x.Insert(LineBreakPoint + 1, "\n *     ");
            }
        }
        else if (x.IndexOf(" = function") > 0)
            x = x.Insert(x.IndexOf(" = function"), "\n   ");

        return x + m.Groups[2].Value;
    }

    private void InitializeTypesMap(string MaxApiVersion)
    {
        //Map types referenced in the jQuery API to Closure recognized types
        TypesMap.Add("ajaxSettings", "Object.<string, *>");
        TypesMap.Add("Any", "*");
        TypesMap.Add("Anything", "*");
        TypesMap.Add("Array", "Array");
        TypesMap.Add("attr", "string");
        TypesMap.Add("boolean", "boolean");
        TypesMap.Add("Callback", "function()");
        TypesMap.Add("document", "Document");
        TypesMap.Add("Element", "Element");
        TypesMap.Add("Elements", "Array.<Element>");
        TypesMap.Add("Function", "function()");
        TypesMap.Add("string", "string");
        TypesMap.Add("Integer", "number");
        TypesMap.Add("Map", "Object.<string,*>");
        TypesMap.Add("Map, String", "Object.<string,*>|string");
        TypesMap.Add("next", "function()");
        TypesMap.Add("Number", "number");
        TypesMap.Add("Number/String", "number|string");
        TypesMap.Add("Object", "Object");
        TypesMap.Add("Options", "Object.<string,*>");
        TypesMap.Add("Selector", "jQuerySelector");
        TypesMap.Add("HTML", "string");
        TypesMap.Add("jQuery", "jQuery");
        TypesMap.Add("undefined", "undefined");
        TypesMap.Add("XMLHttpRequest", "XMLHttpRequest");
        TypesMap.Add("xhr", "XMLHttpRequest");
        TypesMap.Add("eventObject", "jQuery.event");
        TypesMap.Add("index", "number");
        TypesMap.Add("data", "string");
        TypesMap.Add("textStatus", "string");
        TypesMap.Add("width", "number");
        TypesMap.Add("height", "number");
        TypesMap.Add("elementOfArray", "*");
        TypesMap.Add("indexInArray", "number");
        TypesMap.Add("valueOfElement", "*");
        TypesMap.Add("coords", "{top:number,left:number}");
        TypesMap.Add("oldhtml", "string");
        TypesMap.Add("domElement", "Element");
        TypesMap.Add("class", "string");
        TypesMap.Add("thrownError", "*");
        TypesMap.Add("value", "*");
        TypesMap.Add("text", "string");
        TypesMap.Add("ajaxOptions", "Object.<string, *>");
        TypesMap.Add("event", "jQuery.event");
        TypesMap.Add("responseText", "string");
        TypesMap.Add("Deferred", "jQuery.deferred");
        TypesMap.Add("XMLDocument", "Document");
        TypesMap.Add("jqXHR", "jQuery.jqXHR");
        TypesMap.Add("Promise", "jQuery.Promise");
        TypesMap.Add("Array, Object", "Object.<string, *>|Array.<Object.<string, *>>");
        TypesMap.Add("currentClass", "String");
        TypesMap.Add("oldPropertyValue", "String");
        TypesMap.Add("originalOptions", "Object.<string, *>");
        TypesMap.Add("indexOrKey", "string|number");
    }

    private void InitializeObjectTypes(string MaxApiVersion)
    {
        //Functions with Object return types
        ObjectTypes.Add("jQuery.support", "Object.<string, *>");
        ObjectTypes.Add("jQuery.cssHooks", "Object.<string, *>");
        ObjectTypes.Add("jQuery.event.prototype.result", "*");
        ObjectTypes.Add("jQuery.prototype.offset", "{left:number,top:number}");
        ObjectTypes.Add("jQuery.prototype.position", "{left:number,top:number}");
        ObjectTypes.Add("jQuery.prototype.isArray", "*");
        ObjectTypes.Add("jQuery.prototype.isFunction", "*");
        ObjectTypes.Add("jQuery.parseJSON", "Object.<string, *>");
        ObjectTypes.Add("jQuery.data", "*");
        ObjectTypes.Add("jQuery.prototype.data", "*");
        ObjectTypes.Add("jQuery.prototype.triggerhandler", "*");
        ObjectTypes.Add("jQuery.map", "Array.<*>|Object.<string,*>");

        //Parameters names of type object
        ObjectTypes.Add("(args)", "*");
        ObjectTypes.Add("(coordinates)", "{left:number,top:number}");
        ObjectTypes.Add("(eventData)", "Object.<string, *>");
        ObjectTypes.Add("(event)", "jQuery.event");
        ObjectTypes.Add("(events)", "Object.<string, function(jQuery.event=)>");
        ObjectTypes.Add("(jQuery object)", "jQuery");
        ObjectTypes.Add("(obj)", "*");
        ObjectTypes.Add("(props)", "Object.<string, (string|function(jQuery.event=))>");
        ObjectTypes.Add("(value)", "*");

        //Overrides for specific function/parameter pairs
        ObjectTypes.Add("jQuery.prototype.data(obj)", "Object.<string, *>");
        ObjectTypes.Add("jQuery(jQuery object)", "Object|jQuery");
        ObjectTypes.Add("jQuery.map(arrayOrObject)", "Object.<string, *>");
        ObjectTypes.Add("jQuery.param(obj)", "Object.<string, *>|Array.<Object.<string, *>>");
    }

    private void InitializeArrayTypes(string MaxApiVersion)
    {
        //Functions with Array return types
        ArrayTypes.Add("jQuery.prototype.closest", "Array.<Element>");
        ArrayTypes.Add("jQuery.prototype.get", "Array.<Element>");
        ArrayTypes.Add("jQuery.grep", "Array.<*>");
        ArrayTypes.Add("jQuery.makeArray", "Array.<*>");
        ArrayTypes.Add("jQuery.map", "Array.<*>");
        ArrayTypes.Add("jQuery.merge", "Array.<*>");
        ArrayTypes.Add("jQuery.queue", "Array.<Element>");
        ArrayTypes.Add("jQuery.unique", "Array.<Element>");
        ArrayTypes.Add("jQuery.prototype.queue", "Array.<Element>");
        ArrayTypes.Add("jQuery.prototype.serializeArray", "Array.<Object.<string, *>>");
        ArrayTypes.Add("jQuery.prototype.toArray", "Array.<Element>");
        ArrayTypes.Add("jQuery.prototype.val", "Array.<string>");

        //Parameters names of type Array
        ArrayTypes.Add("(args)", "Array.<*>");
        ArrayTypes.Add("(arguments)", "Array.<*>");
        ArrayTypes.Add("(array)", "Array.<*>");
        ArrayTypes.Add("(content)", "Array.<Element>");
        ArrayTypes.Add("(elementArray)", "Array.<Element>");
        ArrayTypes.Add("(elements)", "Array.<Element>");
        ArrayTypes.Add("(extraParameters)", "Array.<*>");
        ArrayTypes.Add("(newQueue)", "Array.<function()>");
        ArrayTypes.Add("(selectors)", "string|Array.<string>");

        //Overrides for specific function/parameter pairs
        ArrayTypes.Add("jQuery.unique(array)", "Array.<Element>");
        ArrayTypes.Add("jQuery.merge(first)", "Array.<*>");
        ArrayTypes.Add("jQuery.merge(second)", "Array.<*>");
        ArrayTypes.Add("jQuery.map(arrayOrObject)", "Array.<*>");
    }

    private void InitializeEntries(string MaxApiVersion)
    {
        // Typedef the standard jQuery selector for convenience and readability
        Entries.Add("$_", @"/** @typedef {(Window|Document|Element|Array.<Element>|string|jQuery)} */
var jQuerySelector;
");
        // Add the $ alias
        Entries.Add("jQuery$", @"/**
 * @constructor
 * @extends {jQuery}
 * @param {(jQuerySelector|Element|Array.<Element>|Object|jQuery|string|
 *     function())} arg1
 * @param {(Element|jQuery|Document|
 *     Object.<string, (string|function(jQuery.event=))>)=} arg2
 * @return {jQuery}
 */
function $(arg1, arg2) {};
");

        //Add subtype definitions or override specific function anotations
        Entries.Add("jQuery.event", @"/**
 * @constructor
 * @param {string} eventType
 */
jQuery.event = function(eventType) {};
");
        Entries.Add("jQuery.event$", @"/**
 * @constructor
 * @extends {jQuery.event}
 * @param {string} eventType
 */
$.event = function(eventType) {};
");

        Entries.Add("jQuery.event.prototype.props", @"/**
 * @type {Object.<string, *>}
 */
jQuery.event.prototype.props;
");
        Entries.Add("jQuery.event.prototype.originalEvent", @"/**
 * @type {Event}
 */
jQuery.event.prototype.originalEvent;
");
        Entries.Add("jQuery.Event", @"/**
 * @constructor
 * @extends {jQuery.event}
 * @param {string} eventType
 * @return {jQuery.Event}
 */
jQuery.Event = function(eventType) {};
");

        Entries.Add("jQuery.Event$", @"/**
 * @constructor
 * @extends {jQuery.event}
 * @param {string} eventType
 * @return {$.Event}
 */
$.Event = function(eventType) {};
");

        Entries.Add("jQuery.fx", @"/** @const */
jQuery.fx = {};
");
        Entries.Add("jQuery.fx$", @"/** @const */
$.fx = {};
");
        Entries.Add("jQuery.fn", @"/** @see http://docs.jquery.com/Plugins/Authoring */
jQuery.fn;
");
        Entries.Add("jQuery.fn$", @"/** @see http://docs.jquery.com/Plugins/Authoring */
$.fn;
");

        // The annotations for the extend functions cannot be properly deduced from the API Xml (var_args, optional) so hard code them
        Entries.Add("jQuery.prototype.extend", @"/**
 * @param {(Object|boolean)} arg1
 * @param {...*} var_args
 * @return {Object}
 */
jQuery.prototype.extend = function(arg1, var_args) {};
");
        Entries.Add("jQuery.extend", @"/**
 * @param {(Object|boolean)} arg1
 * @param {...*} var_args
 * @return {Object}
 */
jQuery.extend = function(arg1, var_args) {};
");
        Entries.Add("jQuery.extend$", @"/**
 * @param {(Object|boolean)} arg1
 * @param {...*} var_args
 * @return {Object}
 */
$.extend = function(arg1, var_args) {};
");

        //Added to force a skip of the API pull
        Entries.Add("jQuery.browser.version", null);

        // Browser flags are not defined in the api xml
        Entries.Add("jQuery.browser", @"/** @type {jQuery.browser_} */
jQuery.browser;
");
        Entries.Add("jQuery.browser$", @"/** @type {jQuery.browser_} */
$.browser;
");
        Entries.Add("jQuery.browser_", @"/**
 * This type is used interally to define the jQuery.browser subtype
 * @constructor
 * @private
 */
jQuery.browser_ = function() {};
");
        Entries.Add("jQuery.browser_.prototype.webkit", @"/**
 * @type {boolean}
 * @const
 */
jQuery.browser_.prototype.webkit;
");
        Entries.Add("jQuery.browser_.prototype.safari", @"/**
 * @type {boolean}
 * @const
 * @deprecated
 */
jQuery.browser_.prototype.safari;
");
        Entries.Add("jQuery.browser_.prototype.opera", @"/**
 * @type {boolean}
 * @const
 */
jQuery.browser_.prototype.opera;
");
        Entries.Add("jQuery.browser_.prototype.msie", @"/**
 * @type {boolean}
 * @const
 */
jQuery.browser_.prototype.msie;
");
        Entries.Add("jQuery.browser_.prototype.mozilla", @"/**
 * @type {boolean}
 * @const
 */
jQuery.browser_.prototype.mozilla;
");
        Entries.Add("jQuery.browser_.prototype.version", @"/**
 * @type {string}
 * @const
 */
jQuery.browser_.prototype.version;
");

        // jQuery.support flags are not defined in the api xml
        Entries.Add("jQuery.support", @"/** @type {jQuery.support_} */
jQuery.support;
");
        Entries.Add("jQuery.support$", @"/** @type {jQuery.support_} */
$.support;
");
        Entries.Add("jQuery.support_", @"/**
 * This type is used interally to define the jQuery.support subtype
 * @constructor
 * @private
 */
jQuery.support_ = function() {};
");
        Entries.Add("jQuery.support_.prototype.boxModel", @"/** @type {boolean} */
jQuery.support_.prototype.boxModel;
");
        Entries.Add("jQuery.support_.prototype.changeBubbles", @"/** @type {boolean} */
jQuery.support_.prototype.changeBubbles;
");
        Entries.Add("jQuery.support_.prototype.cssFloat", @"/** @type {boolean} */
jQuery.support_.prototype.cssFloat;
");
        Entries.Add("jQuery.support_.prototype.hrefNormalized", @"/** @type {boolean} */
jQuery.support_.prototype.hrefNormalized;
");
        Entries.Add("jQuery.support_.prototype.htmlSerialize", @"/** @type {boolean} */
jQuery.support_.prototype.htmlSerialize;
");
        Entries.Add("jQuery.support_.prototype.leadingWhitespace", @"/** @type {boolean} */
jQuery.support_.prototype.leadingWhitespace;
");
        Entries.Add("jQuery.support_.prototype.noCloneEvent", @"/** @type {boolean} */
jQuery.support_.prototype.noCloneEvent;
");
        Entries.Add("jQuery.support_.prototype.opacity", @"/** @type {boolean} */
jQuery.support_.prototype.opacity;
");
        Entries.Add("jQuery.support_.prototype.scriptEval", @"/** @type {boolean} */
jQuery.support_.prototype.scriptEval;
");
        Entries.Add("jQuery.support_.prototype.style", @"/** @type {boolean} */
jQuery.support_.prototype.style;
");
        Entries.Add("jQuery.support_.prototype.submitBubbles", @"/** @type {boolean} */
jQuery.support_.prototype.submitBubbles;
");
        Entries.Add("jQuery.support_.prototype.tbody", @"/** @type {boolean} */
jQuery.support_.prototype.tbody;
");

        if (MaxApiVersion.CompareTo("1.5") >= 0)
        {
            //jQuery 1.5 types not in API XML
            Entries.Add("jQuery.jqXHR", @"/**
 * @constructor
 * @extends {XMLHttpRequest}
 * @implements {jQuery.Promise}
 * @private
 * @see http://api.jquery.com/jQuery.ajax/#jqXHR
 */
jQuery.jqXHR = function () {};
");
            Entries.Add("jQuery.jqXHR.prototype.complete", @"/**
 * @param {function()} callback
 * @return {jQuery.jqXHR}
*/
jQuery.jqXHR.prototype.complete = function (callback) {};
");
            Entries.Add("jQuery.jqXHR.prototype.done", @"/**
 * @override
 * @param {function()} doneCallbacks
 * @return {jQuery.Promise}
 */
jQuery.jqXHR.prototype.done = function(doneCallbacks) {};
");
            Entries.Add("jQuery.jqXHR.prototype.error", @"/**
 * @param {function()} callback
 * @return {jQuery.jqXHR}
*/
jQuery.jqXHR.prototype.error = function (callback) {};
");
            Entries.Add("jQuery.jqXHR.prototype.fail", @"/**
 * @override
 * @param {function()} failCallbacks
 * @return {jQuery.Promise}
 */
jQuery.jqXHR.prototype.fail = function(failCallbacks) {};
");
            Entries.Add("jQuery.jqXHR.prototype.isRejected", @"/**
 * @override
 * @return {boolean}
 * @nosideeffects
 */
jQuery.jqXHR.prototype.isRejected = function() {};
");
            Entries.Add("jQuery.jqXHR.prototype.isResolved", @"/**
 * @override
 * @return {boolean}
 * @nosideeffects
 */
jQuery.jqXHR.prototype.isResolved = function() {};
");
            Entries.Add("jQuery.jqXHR.prototype.success", @"/**
 * @param {function()} callback
 * @return {jQuery.jqXHR}
*/
jQuery.jqXHR.prototype.success = function (callback) {};
");
            Entries.Add("jQuery.jqXHR.prototype.then", @"/**
 * @override
 * @param {function()} doneCallbacks
 * @param {function()} failCallbacks
 * @return {jQuery.Promise}
 */
jQuery.jqXHR.prototype.then = function(doneCallbacks, failCallbacks) {};
");
            Entries.Add("jQuery.jqXHR.prototype.onreadystatechange", @"/**
 * @override
 * @deprecated
 */
jQuery.jqXHR.prototype.onreadystatechange = function (callback) {};
");
            Entries.Add("jQuery.Promise", @"/**
 * @interface
 * @private
 * @see http://api.jquery.com/Types/#Promise
 */
jQuery.Promise = function () {};
");
            Entries.Add("jQuery.Promise.prototype.done", @"/**
 * @param {function()} doneCallbacks
 * @return {jQuery.Promise}
 */
jQuery.Promise.prototype.done = function(doneCallbacks) {};
");
            Entries.Add("jQuery.Promise.prototype.fail", @"/**
 * @param {function()} failCallbacks
 * @return {jQuery.Promise}
 */
jQuery.Promise.prototype.fail = function(failCallbacks) {};
");
            Entries.Add("jQuery.Promise.prototype.isResolved", @"/**
 * @return {boolean}
 * @nosideeffects
 */
jQuery.Promise.prototype.isResolved = function() {};
");
            Entries.Add("jQuery.Promise.prototype.isRejected", @"/**
 * @return {boolean}
 * @nosideeffects
 */
jQuery.Promise.prototype.isRejected = function() {};
");
            Entries.Add("jQuery.Promise.prototype.then", @"/**
 * @param {function()} doneCallbacks
 * @param {function()} failCallbacks
 * @return {jQuery.Promise}
 */
jQuery.Promise.prototype.then = function(doneCallbacks, failCallbacks) {};
");
            Entries.Add("jQuery.deferred", @"/**
 * @constructor
 * @param {function()=} opt_fn
 * @see http://api.jquery.com/category/deferred-object/
 */
jQuery.deferred = function(opt_fn) {};
");
            Entries.Add("jQuery.deferred$", @"/**
 * @constructor
 * @extends {jQuery.deferred}
 * @param {function()=} opt_fn
 * @see http://api.jquery.com/category/deferred-object/
 */
$.deferred = function(opt_fn) {};
");
            Entries.Add("jQuery.Deferred", @"/**
 * @constructor
 * @extends {jQuery.deferred}
 * @param {function()=} opt_fn
 * @return {jQuery.Deferred}
 */
jQuery.Deferred = function(opt_fn) {};
");
            Entries.Add("jQuery.Deferred$", @"/**
 * @constructor
 * @extends {jQuery.deferred}
 * @param {function()=} opt_fn
 * @return {jQuery.Deferred}
 */
$.Deferred = function(opt_fn) {};
");
        }
    }

    private void InitializeNoSideEffectMaps(string MaxApiVersion)
    {
        NoSideEffectCategories.Add("Miscellaneous Traversing");
        NoSideEffectCategories.Add("Tree Traversal");
        NoSideEffectCategories.Add("Filtering");

        NoSideEffectPrefixes.Add("is");
        NoSideEffectPrefixes.Add("has");

        NoSideEffectMethods.Add("jQuery.prototype.get", "jQuery.prototype.get");
        NoSideEffectMethods.Add("jQuery.prototype.innerHeight", "jQuery.prototype.innerHeight");
        NoSideEffectMethods.Add("jQuery.prototype.innerWidth", "jQuery.prototype.innerWidth");
        NoSideEffectMethods.Add("jQuery.inArray", "jQuery.inArray");
        NoSideEffectMethods.Add("jQuery.noop", "jQuery.noop");
        NoSideEffectMethods.Add("jQuery.now", "jQuery.now");
        NoSideEffectMethods.Add("jQuery.sub", "jQuery.sub");
        NoSideEffectMethods.Add("jQuery.trim", "jQuery.trim");
        NoSideEffectMethods.Add("jQuery.type", "jQuery.type");
        NoSideEffectMethods.Add("jQuery.prototype.outerHeight", "jQuery.prototype.outerHeight");
        NoSideEffectMethods.Add("jQuery.prototype.outerWidth", "jQuery.prototype.outerWidth");
        NoSideEffectMethods.Add("jQuery.prototype.position", "jQuery.prototype.position");
        NoSideEffectMethods.Add("jQuery.prototype.toArray", "jQuery.prototype.toArray");
        NoSideEffectMethods.Add("jQuery.prototype.serialize", "jQuery.prototype.serialize");
        NoSideEffectMethods.Add("jQuery.prototype.serializeArray", "jQuery.prototype.serializeArray");
        NoSideEffectMethods.Add("jQuery.prototype.size", "jQuery.prototype.size");
    }
}

public class jQueryEntry : IComparable<jQueryEntry>
{
    public string EntryName { get; set; }
    public string Value { get; set; }
    public jQueryEntry(string EntryName, string Value)
    {
        this.EntryName = EntryName;
        this.Value = Value;
    }

    public int CompareTo(jQueryEntry y)
    {
        int Value = this.EntryName.Replace(".prototype", "").CompareTo(y.EntryName.Replace(".prototype", ""));
        if (Value == 0)
            return this.EntryName.CompareTo(y.EntryName);

        return Value;
    }
}

public class jQueryEntryArgument
{
    public int Position { get; set; }
    public string Name { get; set; }
    public List<string> Types
    {
        get { return Types_; }
    }
    private List<string> Types_ = new List<string>();
    public bool Optional { get; set; }

    public jQueryEntryArgument(int Position)
    {
        this.Position = Position;
    }

    public void UnionTypes(List<string> OtherTypes)
    {
        foreach (string ArgType in OtherTypes)
        {
            AddType(ArgType);
        }
    }

    public void AddType(string ArgType)
    {
        if (Types.Count == 1 && (Types[0] == "*" || Types[0] == "...*"))
            return;

        if (ArgType == "...*" || ArgType == "*")
        {
            Types.Clear();
            Types.Add(ArgType);
        }
        else if (!Types.Contains(ArgType))
            Types.Add(ArgType);
    }
}